import sys
from pathlib import Path
from typing import List, Dict
from datetime import datetime

# 添加项目根目录到Python路径
sys.path.append(str(Path(__file__).parent.parent.parent))

from fmpxx.updatedb.database import FMPDatabase
from fmpxx.quote import Quote

def update_historical_prices(api_key: str, symbol: str, db_path: str = "fmp_data.db", period: int = 1) -> int:
    """
    更新指定股票的日线历史数据
    
    Args:
        api_key (str): FMP API密钥
        symbol (str): 股票代码
        db_path (str): 数据库路径，默认为"fmp_data.db"
        period (int): 获取的年数，默认为1年
        
    Returns:
        int: 更新的记录数
    """
    quote = Quote(api_key)
    df = quote.get_his_daily(symbol, period)
    print(df)
    if df.empty:
        return 0
        
    db = FMPDatabase(db_path)
    updated_count = 0
    
    with db as db:
        # 添加symbol列
        df['symbol'] = symbol
        # 将DataFrame直接写入数据库
        df.to_sql(
            'historical_prices',
            db.conn,
            if_exists='append',
            index=False,
            method='multi'
        )
        updated_count = len(df)
        db.conn.commit()
        
    return updated_count
    
def update_all_prices(api_key: str, symbols: List[str], db_path: str = "fmp_data.db", period: int = 1) -> Dict[str, int]:
    """
    批量更新多个股票的日线历史数据
    
    Args:
        api_key (str): FMP API密钥
        symbols (List[str]): 股票代码列表
        db_path (str): 数据库路径，默认为"fmp_data.db"
        period (int): 获取的年数，默认为1年
        
    Returns:
        Dict[str, int]: 每个股票更新的记录数
    """
    results = {}
    for symbol in symbols:
        count = update_historical_prices(api_key, symbol, db_path, period)
        results[symbol] = count
    return results


if __name__ == "__main__":
    update_historical_prices('aapl',1)



"""Custom exceptions for the FMP API client."""

class FMPAPIError(Exception):
    """Base exception for FMP API errors."""
    pass

class InvalidAPIKeyError(FMPAPIError):
    """Raised when the provided API key is invalid or unauthorized (HTTP 401)."""
    pass

class SymbolNotFoundError(FMPAPIError):
    """Raised when a requested symbol or resource is not found (HTTP 404)."""
    pass

class RateLimitExceededError(FMPAPIError):
    """Raised when the API rate limit is exceeded (HTTP 429)."""
    pass

class FMPConnectionError(FMPAPIError):
    """Raised for network-related errors during API requests."""
    pass



import sys
import os
import dotenv
import pandas as pd

# Set pandas display options for better output in console
pd.set_option('display.max_columns', None)
pd.set_option('display.width', 1000)

# 获取项目根目录的绝对路径
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
# 将项目根目录添加到 Python 模块搜索路径
sys.path.append(project_root)

# 然后再导入你的模块
from fmpxx import FMPClient
from fmpxx.exceptions import FMPAPIError, InvalidAPIKeyError, SymbolNotFoundError, RateLimitExceededError, FMPConnectionError

dotenv.load_dotenv()

API_Key = os.getenv("FMP_KEY")
ticker = 'AAPL' # Changed to AAPL for consistency

if not API_Key:
    print("Error: FMP_KEY environment variable not set. Please set it in your .env file or environment.")
    sys.exit(1)

client = FMPClient(api_key=API_Key, output_format='pandas')

print("\n--- Testing Financials ---")
try:
    # Test income statement
    print(f"Fetching income statement for {ticker}...")
    income_statement = client.financials.get_financials(statement='income', symbol=ticker, limit=1, period='quarter')
    print("Income Statement (first row):\n", income_statement.head(1))

    # Test balance sheet
    print(f"Fetching balance sheet for {ticker}...")
    balance_sheet = client.financials.get_financials(statement='balance', symbol=ticker, limit=1, period='quarter')
    print("Balance Sheet (first row):\n", balance_sheet.head(1))

    # Test cash flow statement
    print(f"Fetching cash flow statement for {ticker}...")
    cash_flow_statement = client.financials.get_financials(statement='cash', symbol=ticker, limit=1, period='quarter')
    print("Cash Flow Statement (first row):\n", cash_flow_statement.head(1))

    # Test merged financials
    print(f"Fetching merged financials for {ticker}...")
    merged_financials = client.financials.get_merged_financials(ticker, limit=1, period='quarter')
    print("Merged Financials (first row):\n", merged_financials.head(1))

except InvalidAPIKeyError:
    print("Test Failed: Invalid API Key. Please check your FMP_KEY.")
except SymbolNotFoundError as e:
    print(f"Test Failed: Symbol not found. {e}")
except RateLimitExceededError:
    print("Test Failed: Rate limit exceeded. Please wait and try again.")
except FMPConnectionError as e:
    print(f"Test Failed: Connection error. {e}")
except FMPAPIError as e:
    print(f"Test Failed: An FMP API error occurred. {e}")
except Exception as e:
    print(f"An unexpected error occurred during financials test: {e}")

print("\n--- Testing Stocks Data ---")
try:
    # Test historical price full
    print(f"Fetching historical daily prices for {ticker}...")
    historical_prices = client.stocks.historical_price_full(ticker, from_date="2023-01-01", to_date="2023-01-05")
    print("Historical Prices (first 5 rows):\n", historical_prices.head())

    # Test stock list (limit to first 5 for brevity)
    print("Fetching stock list (first 5 entries)...")
    stock_list = client.stocks.stock_list()
    print("Stock List (first 5 rows):\n", stock_list.head())

    # Test quote
    print(f"Fetching quote for {ticker}...")
    quote = client.stocks.quote(ticker)
    print("Quote:\n", quote)

    # Test search
    print(f"Searching for 'Apple'...")
    search_results = client.stocks.search(query='Apple', limit=2)
    print("Search Results (first 2 rows):\n", search_results.head())

except InvalidAPIKeyError:
    print("Test Failed: Invalid API Key. Please check your FMP_KEY.")
except SymbolNotFoundError as e:
    print(f"Test Failed: Symbol not found. {e}")
except RateLimitExceededError:
    print("Test Failed: Rate limit exceeded. Please wait and try again.")
except FMPConnectionError as e:
    print(f"Test Failed: Connection error. {e}")
except FMPAPIError as e:
    print(f"Test Failed: An FMP API error occurred. {e}")
except Exception as e:
    print(f"An unexpected error occurred during stocks test: {e}")

# Old code commented out
# from fmpxx import FMPClient, Quote
# from fmpxx.financials import Financials
# import os
# import dotenv
# dotenv.load_dotenv()
# API_Key = os.getenv("FMP_KEY")
# ticker = 'aapl'
# fin = Financials(API_Key)
# # print(fin.get_income_statement(ticker))
# # print(fin.get_earnings_his(ticker))
# print(fin.get_earnings_transcript(ticker,period=3))
# # print(client.get_earnings_his('NVDA', period=1))
# # quote = Quote(api_key=API_Key)
# # print(quote.get_his_daily(symbol='aapl', period=1))
# # print(quote.get_simple_quote('aapl'))
# # print(quote.get_full_quote('aapl'))



import requests
import pandas as pd
from .exceptions import FMPAPIError, InvalidAPIKeyError, SymbolNotFoundError, RateLimitExceededError, FMPConnectionError

class _BaseClient:
    BASE_URL = "https://financialmodelingprep.com/api/v3/"

    def __init__(self, api_key: str, timeout: int = 10, output_format: str = 'json'):
        if not api_key:
            raise ValueError("API key is required.")
        self.api_key = api_key
        self.timeout = timeout
        self.output_format = output_format
        self.session = requests.Session()
        self.session.params.update({'apikey': self.api_key})

    def _make_request(self, endpoint: str, params: dict = None):
        url = f"{self.BASE_URL}{endpoint}"
        full_params = params.copy() if params else {}

        try:
            response = self.session.get(url, params=full_params, timeout=self.timeout)
            response.raise_for_status()  # Raises HTTPError for bad responses (4xx or 5xx)
        except requests.exceptions.HTTPError as e:
            if response.status_code == 401:
                raise InvalidAPIKeyError("Invalid API key. Please check your API key.") from e
            elif response.status_code == 404:
                raise SymbolNotFoundError(f"Resource not found for endpoint: {endpoint}. Details: {response.text}") from e
            elif response.status_code == 429:
                raise RateLimitExceededError("API rate limit exceeded. Please wait and try again.") from e
            else:
                raise FMPAPIError(f"FMP API error: {response.status_code} - {response.text}") from e
        except requests.exceptions.ConnectionError as e:
            raise FMPConnectionError(f"Network connection error: {e}") from e
        except requests.exceptions.Timeout as e:
            raise FMPConnectionError(f"Request timed out after {self.timeout} seconds: {e}") from e
        except requests.exceptions.RequestException as e:
            raise FMPAPIError(f"An unexpected request error occurred: {e}") from e

        try:
            data = response.json()
        except ValueError as e:
            raise FMPAPIError(f"Failed to decode JSON response: {e}. Response content: {response.text}") from e

        return data

    def _process_response(self, data):
        if self.output_format == 'pandas' and isinstance(data, list):
            if not data:
                return pd.DataFrame() # Return empty DataFrame for empty list
            try:
                df = pd.DataFrame(data)
                # Drop specific columns as requested
                df = df.drop(columns=["link", "finalLink"], errors="ignore")
                return df
            except Exception as e:
                # Fallback to JSON if pandas conversion fails for some reason
                print(f"Warning: Could not convert to Pandas DataFrame. Returning JSON. Error: {e}")
                return data
        return data



import sys
import os
import dotenv
import pandas as pd

# Set pandas display options for better output in console
# 不换行显示配置
pd.set_option('display.width', 1000)          # 宽度足够大
pd.set_option('display.max_columns', None)     # 显示所有列
pd.set_option('display.max_colwidth', 30)      # 限制单列宽度
pd.set_option('display.expand_frame_repr', False)  # 禁用折叠

# 获取项目根目录的绝对路径
project_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
# 将项目根目录添加到 Python 模块搜索路径
sys.path.append(project_root)

# 然后再导入你的模块
from fmpxx import FMPClient
from fmpxx.exceptions import FMPAPIError, InvalidAPIKeyError, SymbolNotFoundError, RateLimitExceededError, FMPConnectionError

dotenv.load_dotenv()

API_Key = os.getenv("FMP_KEY")
ticker = 'AAPL' # Changed to AAPL for consistency

if not API_Key:
    print("Error: FMP_KEY environment variable not set. Please set it in your .env file or environment.")
    sys.exit(1)

client = FMPClient(api_key=API_Key, output_format='pandas')

print("\n--- Testing Financials ---")

# Test income statement
print(f"Fetching income statement for {ticker}...")
income_statement = client.financials.get_financials(statement='income',symbol=ticker, limit=4)
print("Income Statement (first row):\n", income_statement)


import sqlite3
from pathlib import Path
from typing import Optional

class FMPDatabase:
    """管理FMP数据的SQLite数据库"""
    
    def __init__(self, db_path: str = "fmp_data.db"):
        self.db_path = db_path
        self.conn: Optional[sqlite3.Connection] = None
        
    def connect(self) -> sqlite3.Connection:
        """连接到数据库，如果不存在则创建"""
        Path(self.db_path).parent.mkdir(parents=True, exist_ok=True)
        self.conn = sqlite3.connect(self.db_path)
        return self.conn
        
    def initialize(self) -> None:
        """初始化数据库表结构"""
        with self.connect() as conn:
            # 创建日线历史数据表
            conn.execute("""
                CREATE TABLE IF NOT EXISTS historical_prices (
                    symbol TEXT NOT NULL,
                    date TEXT NOT NULL,
                    open REAL,
                    high REAL,
                    low REAL,
                    close REAL,
                    volume INTEGER,
                    PRIMARY KEY (symbol, date)
                )
            """)
            
            # 创建财务数据表
            conn.execute("""
                CREATE TABLE IF NOT EXISTS financials (
                    symbol TEXT NOT NULL,
                    date TEXT NOT NULL,
                    revenue REAL,
                    net_income REAL,
                    total_assets REAL,
                    total_liabilities REAL,
                    operating_cash_flow REAL,
                    PRIMARY KEY (symbol, date)
                )
            """)
            
            # 创建财报电话会议表
            conn.execute("""
                CREATE TABLE IF NOT EXISTS earnings_transcripts (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    symbol TEXT NOT NULL,
                    date TEXT NOT NULL,
                    quarter INTEGER,
                    year INTEGER,
                    content TEXT,
                    UNIQUE(symbol, date)
                )
            """)
            
            # 创建索引
            conn.execute("CREATE INDEX IF NOT EXISTS idx_symbol ON historical_prices(symbol)")
            conn.execute("CREATE INDEX IF NOT EXISTS idx_financials_symbol ON financials(symbol)")
            conn.execute("CREATE INDEX IF NOT EXISTS idx_transcripts_symbol ON earnings_transcripts(symbol)")
            
    def close(self) -> None:
        """关闭数据库连接"""
        if self.conn:
            self.conn.close()
            self.conn = None
            
    def __enter__(self):
        self.connect()
        return self
        
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



from .base import _BaseClient
from .financials import Financials
from .stocks import Stocks

class FMPClient(_BaseClient):
    """Main client for interacting with the Financial Modeling Prep (FMP) API.

    This class provides a user-friendly and consistent interface to access
    various financial data endpoints offered by FMP.

    Args:
        api_key (str): Your FMP API key.
        timeout (int, optional): Request timeout in seconds. Defaults to 10.
        output_format (str, optional): Desired output format ('json' or 'pandas').
                                       Defaults to 'json'.

    Attributes:
        financials (Financials): Access to company fundamental data.
        stocks (Stocks): Access to stock market data.
    """

    def __init__(self, api_key: str, timeout: int = 10, output_format: str = 'json'):
        super().__init__(api_key, timeout, output_format)

        # Initialize categorized API modules
        self.financials = Financials(api_key, timeout, output_format)
        self.stocks = Stocks(api_key, timeout, output_format)

    # You can add more categorized properties here as you implement more modules
    # For example:
    # @property
    # def forex(self):
    #     return Forex(self.api_key, self.timeout, self.output_format)

    # @property
    # def cryptocurrencies(self):
    #     return Cryptocurrencies(self.api_key, self.timeout, self.output_format)


from .base import _BaseClient
import pandas as pd

class Stocks(_BaseClient):
    """Client for FMP Stock API endpoints."""

    def __init__(self, api_key: str, timeout: int = 10, output_format: str = 'json'):
        super().__init__(api_key, timeout, output_format)

    def historical_price_full(self, symbol: str, series_type: str = None, from_date: str = None, to_date: str = None):
        """
        Get full historical daily prices for a given symbol.

        Args:
            symbol (str): Stock ticker symbol (e.g., 'AAPL').
            series_type (str, optional): Type of series (e.g., 'line').
            from_date (str, optional): Start date in YYYY-MM-DD format.
            to_date (str, optional): End date in YYYY-MM-DD format.

        Returns:
            list or pandas.DataFrame: Historical price data.
        """
        endpoint = f"historical-price-full/{symbol}"
        params = {}
        if series_type: params['serietype'] = series_type
        if from_date: params['from'] = from_date
        if to_date: params['to'] = to_date

        data = self._make_request(endpoint, params)
        if data and 'historical' in data:
            return self._process_response(data['historical'])
        return self._process_response(data)

    def daily_prices(self, symbol: str, from_date: str = None, to_date: str = None):
        """
        Get historical daily prices for a given symbol (line series).

        Args:
            symbol (str): Stock ticker symbol (e.g., 'AAPL').
            from_date (str, optional): Start date in YYYY-MM-DD format.
            to_date (str, optional): End date in YYYY-MM-DD format.

        Returns:
            list or pandas.DataFrame: Daily price data.
        """
        return self.historical_price_full(symbol, series_type='line', from_date=from_date, to_date=to_date)

    def stock_list(self):
        """
        Get a list of all available stocks.

        Returns:
            list or pandas.DataFrame: List of stocks.
        """
        endpoint = "stock/list"
        data = self._make_request(endpoint)
        return self._process_response(data)

    def quote(self, symbol: str):
        """
        Get real-time quote for a given symbol.

        Args:
            symbol (str): Stock ticker symbol (e.g., 'AAPL').

        Returns:
            list or pandas.DataFrame: Quote data.
        """
        endpoint = f"quote/{symbol}"
        data = self._make_request(endpoint)
        return self._process_response(data)

    def search(self, query: str, exchange: str = None, limit: int = 10):
        """
        Search for companies by name or symbol.

        Args:
            query (str): Search query.
            exchange (str, optional): Filter by exchange.
            limit (int): Number of results to return. Defaults to 10.

        Returns:
            list or pandas.DataFrame: Search results.
        """
        endpoint = "search"
        params = {'query': query, 'limit': limit}
        if exchange: params['exchange'] = exchange
        data = self._make_request(endpoint, params)
        return self._process_response(data)



from typing import List, Dict
from datetime import datetime
from .database import FMPDatabase
from fmpxx.financials import Financials

def update_earnings_transcripts(api_key: str, symbol: str, db_path: str = "fmp_data.db", period: int = 3) -> int:
    """
    更新指定股票的财报电话会议记录
    
    Args:
        api_key (str): FMP API密钥
        symbol (str): 股票代码
        db_path (str): 数据库路径，默认为"fmp_data.db"
        period (int): 获取的年数，默认为3年
        
    Returns:
        int: 更新的记录数
    """
    financials = Financials(api_key)
    transcripts = financials.get_earnings_transcript(symbol, period)
    if not transcripts:
        return 0
        
    db = FMPDatabase(db_path)
    updated_count = 0
    
    with db as db:
        cursor = db.conn.cursor()
        
        for transcript in transcripts:
            try:
                # 解析日期
                date = datetime.strptime(transcript["date"], "%Y-%m-%d").strftime("%Y-%m-%d")
                quarter = transcript.get("quarter")
                year = transcript.get("year")
                content = transcript.get("content", "")
                
                cursor.execute("""
                    INSERT OR REPLACE INTO earnings_transcripts 
                    (symbol, date, quarter, year, content)
                    VALUES (?, ?, ?, ?, ?)
                """, (symbol, date, quarter, year, content))
                
                updated_count += 1
            except Exception as e:
                print(f"Failed to insert transcript for {symbol} on {transcript['date']}: {e}")
                
        db.conn.commit()
        
    return updated_count
    
def update_all_transcripts(api_key: str, symbols: List[str], db_path: str = "fmp_data.db", period: int = 3) -> Dict[str, int]:
    """
    批量更新多个股票的财报电话会议记录
    
    Args:
        api_key (str): FMP API密钥
        symbols (List[str]): 股票代码列表
        db_path (str): 数据库路径，默认为"fmp_data.db"
        period (int): 获取的年数，默认为3年
        
    Returns:
        Dict[str, int]: 每个股票更新的记录数
    """
    results = {}
    for symbol in symbols:
        count = update_earnings_transcripts(api_key, symbol, db_path, period)
        results[symbol] = count
    return results



from .base import _BaseClient
import pandas as pd
import logging
from typing import Any, Dict, Optional

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
logger.addHandler(handler)

class Financials(_BaseClient):
    """Client for FMP Company Fundamentals API endpoints."""

    def __init__(self, api_key: str, timeout: int = 10, output_format: str = 'json'):
        super().__init__(api_key, timeout, output_format)

    def get_financials(
        self,
        symbol: str,
        statement: str,
        limit: int = 10,
        period: str = 'quarter',
        **query_params: Dict[str, Any]
    ) -> pd.DataFrame:
        """
        获取指定类型的财务报表数据。

        Args:
            symbol (str): 股票代码
            statement (str): 报表类型，可选值：'income', 'balance', 'cash', 'ratios', 'enterprise-value', 'key-metrics', 'financial-growth'
            limit (int): 限制返回的数据条数。
            period (str): 报表周期，'annual' 或 'quarter'。
            **query_params: 其他查询参数

        Returns:
            pd.DataFrame: 包含财务报表数据的DataFrame

        Raises:
            ValueError: 如果statement参数无效
        """
        # 映射报表类型到API端点
        endpoint_map = {
            "income": "income-statement",
            "balance": "balance-sheet-statement", 
            "cash": "cash-flow-statement"
        }

        if statement not in endpoint_map:
            raise ValueError(f"Invalid statement type: {statement}")

        # Combine direct parameters with query_params
        all_params = {"limit": limit, "period": period}
        all_params.update(query_params)

        # Use self._make_request for consistency
        res = self._make_request(f"{endpoint_map[statement]}/{symbol}", all_params)
        
        if not res:
            logger.warning(f"获取{symbol}的{statement}报表数据失败")
            return pd.DataFrame()

        return self._process_response(res)

    def get_merged_financials(self, symbol: str, limit: int = 40, period: str = 'quarter') -> Optional[pd.DataFrame]:
        """
        合并现金流量表、损益表和资产负债表三张财务报表。

        Args:
            symbol (str): 股票代码

        Returns:
            Optional[pd.DataFrame]: 合并后的财务报表DataFrame，如果数据无效则返回None

        Note:
            - 使用inner join确保三张报表数据完整
            - 处理特殊股票的数据异常
            - 检查数据连续性和完整性
        """
        # 获取三张财务报表
        income = self.get_financials(symbol, statement="income", period=period, limit=limit)
        balance = self.get_financials(symbol, statement="balance", period=period, limit=limit)
        cash = self.get_financials(symbol, statement="cash", period=period, limit=limit)

        # 检查数据完整性
        if income.empty or balance.empty or cash.empty:
            logger.warning(f"{symbol}的财务报表数据不完整")
            return None

        # 定义合并键
        on_list = [
            "cik", "fillingDate", "date", "symbol", 
            "period", "calendarYear", "reportedCurrency"
        ]

        # 合并财务报表
        merged_df = pd.merge(
            pd.merge(income, balance, how="inner", on=on_list),
            cash, how="inner", on=on_list
        )

        # 数据清洗和转换
        merged_df = (
            merged_df.rename(columns={
                "date": "period_date",
                "fillingDate": "date",
                "netIncome_x": "netIncome"
            })
            .sort_values(by=["date", "acceptedDate_x"], ignore_index=True)
            .drop_duplicates(subset=["date"], keep="first", ignore_index=True)
            .assign(period_date=lambda x: pd.to_datetime(x["period_date"]))
        )

        # 处理特殊股票
        special_symbols = [
            "ADT", "ALTR", "ARNC", "BEAM", "CEG", "CSC", "CTLT", "FTV", 
            "HLT", "HPE", "LDOS", "LW", "MMI", "MRNA", "OTIS", "PLL", 
            "S", "TWTR", "VNT"
        ]
        
        if symbol in special_symbols:
            merged_df = merged_df.drop([0]).reset_index(drop=True)

        # 检查数据连续性
        merged_df["verify"] = merged_df["period_date"].diff()
        if symbol == "CSC":
            return None

        # 验证数据周期
        for days in merged_df["verify"].tolist()[1:]:
            if days.days > 150:  # 正常间隔约91天
                logger.warning(f"{symbol}财报异常，数据周期不全")
                return None

        # 最终清理
        return (
            merged_df.drop(columns=["verify"])
            .fillna(value=0)
        )
